use galactica_content::{Ship, SystemObject, SystemObjectHandle};
use galactica_system::{
	data::{self, ShipData},
	phys::PhysSimShipHandle,
};
use log::error;
use rhai::{CustomType, TypeBuilder};
use std::sync::Arc;

use crate::RenderInput;

#[derive(Debug, Clone)]
pub struct ShipState {
	ship: Option<PhysSimShipHandle>,
	input: Arc<RenderInput>,
}

// TODO: remove this
unsafe impl Send for ShipState {}
unsafe impl Sync for ShipState {}

impl ShipState {
	fn get_content(&mut self) -> &Ship {
		let ship = self
			.input
			.phys_img
			.get_ship(self.ship.as_ref().unwrap())
			.unwrap();
		let handle = ship.ship.get_data().get_content();
		self.input.ct.get_ship(handle)
	}

	fn get_data(&mut self) -> &ShipData {
		let ship = self
			.input
			.phys_img
			.get_ship(self.ship.as_ref().unwrap())
			.unwrap();
		ship.ship.get_data()
	}

	fn landed_on(&mut self) -> SystemObjectState {
		let input = self.input.clone();
		match self.get_data().get_state() {
			data::ShipState::Landed { target } => {
				return SystemObjectState {
					input,
					object: Some(*target),
				}
			}
			_ => {
				return SystemObjectState {
					input,
					object: None,
				}
			}
		};
	}
}

impl CustomType for ShipState {
	fn build(mut builder: TypeBuilder<Self>) {
		builder
			.with_name("ShipState")
			.with_fn("is_some", |s: &mut Self| s.ship.is_some())
			.with_fn("is_dead", |s: &mut Self| s.get_data().get_state().is_dead())
			.with_fn("is_landed", |s: &mut Self| {
				s.get_data().get_state().is_landed()
			})
			.with_fn("is_landing", |s: &mut Self| {
				s.get_data().get_state().is_landing()
			})
			.with_fn("is_flying", |s: &mut Self| {
				s.get_data().get_state().is_flying()
			})
			.with_fn("is_unlanding", |s: &mut Self| {
				s.get_data().get_state().is_unlanding()
			})
			.with_fn("is_collapsing", |s: &mut Self| {
				s.get_data().get_state().is_collapsing()
			})
			.with_fn("name", |s: &mut Self| s.get_content().name.clone())
			.with_fn("thumbnail", |s: &mut Self| s.get_content().thumb)
			.with_fn("landed_on", |s: &mut Self| s.landed_on())
			.with_fn("get_shields", |s: &mut Self| s.get_data().get_shields())
			.with_fn("get_total_shields", |s: &mut Self| {
				s.get_data().get_outfits().get_total_shields()
			})
			.with_fn("get_total_hull", |s: &mut Self| s.get_content().hull)
			.with_fn("get_hull", |s: &mut Self| s.get_data().get_hull());
	}
}

#[derive(Debug, Clone)]
pub struct SystemObjectState {
	object: Option<SystemObjectHandle>,
	input: Arc<RenderInput>,
}

// TODO: remove this
unsafe impl Send for SystemObjectState {}
unsafe impl Sync for SystemObjectState {}

impl SystemObjectState {
	fn get_content(&mut self) -> &SystemObject {
		self.input.ct.get_system_object(self.object.unwrap())
	}
}

impl CustomType for SystemObjectState {
	fn build(mut builder: TypeBuilder<Self>) {
		builder
			.with_name("SystemObjectState")
			//
			// Get landable name
			.with_fn("name", |s: &mut Self| {
				s.get_content()
					.name
					.as_ref()
					.map(|x| x.to_string())
					.unwrap_or_else(|| {
						error!("UI called `name()` on a system object which doesn't provide one");
						"".to_string()
					})
			})
			//
			// Get landable description
			.with_fn("desc", |s: &mut Self| {
				s.get_content()
					.desc
					.as_ref()
					.map(|x| x.to_string())
					.unwrap_or_else(|| {
						error!("UI called `name()` on a system object which doesn't provide one");
						"".to_string()
					})
			})
			//
			// Get landable landscape image
			.with_fn("image", |s: &mut Self| {
				let handle = s.get_content().image;
				if let Some(handle) = handle {
					s.input.ct.get_sprite(handle).name.clone()
				} else {
					error!("UI called `image()` on a system object which doesn't provide one");
					"".to_string()
				}
			})
			.with_fn("is_some", |s: &mut Self| s.object.is_some());
	}
}

#[derive(Debug, Clone)]
pub struct State {
	input: Arc<RenderInput>,
}

// TODO: remove this
unsafe impl Send for State {}
unsafe impl Sync for State {}

impl State {
	pub fn new(input: Arc<RenderInput>) -> Self {
		Self { input }
	}

	pub fn player_ship(&mut self) -> ShipState {
		ShipState {
			input: self.input.clone(),
			ship: self.input.player.ship.map(|x| PhysSimShipHandle(x)),
		}
	}
}

impl CustomType for State {
	fn build(mut builder: TypeBuilder<Self>) {
		builder
			.with_name("State")
			.with_fn("player_ship", Self::player_ship);
	}
}
