use anyhow::{Context, Result};
use galactica_content::{resolve_edge_as_edge, Content};
use galactica_system::phys::PhysSimShipHandle;
use galactica_util::rhai_error_to_anyhow;
use glyphon::{cosmic_text::Align, FamilyOwned, FontSystem, Style, Weight};
use log::{debug, error};
use rhai::{Dynamic, Engine, ImmutableString, Scope};
use std::{cell::RefCell, num::NonZeroU32, rc::Rc, sync::Arc};

use super::{
	api::{self, Color, MouseClickEvent, MouseHoverEvent, PlayerShipStateEvent, Rect},
	event::Event,
	util::{RadialBar, Sprite, TextBox},
	UiConfig, UiElement, UiState,
};
use crate::{ui::api::State, RenderInput, RenderState};

pub(crate) struct UiScriptExecutor {
	engine: Engine,
	scope: Scope<'static>,

	pub state: Rc<RefCell<UiState>>,

	last_player_state: u32,
	last_scene: Option<ImmutableString>,
}

impl UiScriptExecutor {
	pub fn new(ct: Arc<Content>, state: &mut RenderState) -> Self {
		let scope = Scope::new();
		let elements = Rc::new(RefCell::new(UiState::new(ct.clone(), state)));

		let mut engine = Engine::new_raw();
		api::register_into_engine(&mut engine);
		Self::register_api(
			ct.clone(),
			state.text_font_system.clone(),
			elements.clone(),
			&mut engine,
		);

		Self {
			engine,
			scope,
			state: elements,
			last_scene: None,
			last_player_state: 0,
		}
	}

	pub fn get_config(&self) -> UiConfig {
		(*self.state).borrow().config.clone()
	}

	/// Change the current scene
	pub fn set_scene(&mut self, input: Arc<RenderInput>) -> Result<()> {
		let current_scene = (*self.state).borrow().get_scene().clone();
		if self.last_scene == current_scene {
			return Ok(());
		}
		self.last_scene = current_scene.clone();

		if current_scene.is_none() {
			return Ok(());
		}

		debug!(
			"switched to {}, running `init()`",
			current_scene.as_ref().unwrap()
		);

		self.scope.clear();

		// Drop this right away, since all script calls borrow elm mutably.
		let mut elm = self.state.borrow_mut();
		elm.clear();
		drop(elm);
		let ct = (*self.state).borrow().ct.clone();

		rhai_error_to_anyhow(
			self.engine.call_fn(
				&mut self.scope,
				ct.get_config()
					.ui_scenes
					.get(current_scene.as_ref().unwrap().as_str())
					.unwrap(),
				"init",
				(State::new(input.clone()),),
			),
		)
		.with_context(|| format!("while running `init()`"))
		.with_context(|| format!("in ui scene `{}`", current_scene.as_ref().unwrap()))?;

		return Ok(());
	}

	/// Draw all ui elements
	pub fn draw(&mut self, state: &mut RenderState, input: Arc<RenderInput>) -> Result<()> {
		let ct = (*self.state).borrow().ct.clone();

		// Initialize start scene if we haven't yet
		if (*self.state).borrow().get_scene().is_none() {
			(*self.state)
				.borrow_mut()
				.set_scene(ImmutableString::from(&ct.get_config().start_ui_scene));
		}
		self.set_scene(input.clone())?;
		let current_scene = (*self.state).borrow().get_scene().clone();

		(*self.state).borrow_mut().step(state, input.clone());

		// Run step() (if it is defined)

		let ast = ct
			.get_config()
			.ui_scenes
			.get(current_scene.as_ref().unwrap().as_str())
			.unwrap();
		if ast.iter_functions().any(|x| x.name == "step") {
			rhai_error_to_anyhow(self.engine.call_fn(
				&mut self.scope,
				ast,
				"step",
				(State::new(input.clone()),),
			))
			.with_context(|| format!("while calling `step()`"))
			.with_context(|| format!("in ui scene `{}`", current_scene.as_ref().unwrap()))?;
		}

		// Send player state change events
		if {
			let player = input.player.ship;
			if let Some(player) = player {
				let ship = input.phys_img.get_ship(&PhysSimShipHandle(player)).unwrap();
				if self.last_player_state == 0
					|| NonZeroU32::new(self.last_player_state).unwrap()
						!= ship.ship.get_data().get_state().as_int()
				{
					self.last_player_state = ship.ship.get_data().get_state().as_int().into();
					true
				} else {
					false
				}
			} else {
				self.last_player_state = 0;
				true
			}
		} {
			rhai_error_to_anyhow(
				self.engine.call_fn(
					&mut self.scope,
					ct.get_config()
						.ui_scenes
						.get(current_scene.as_ref().unwrap().as_str())
						.unwrap(),
					"event",
					(State::new(input.clone()), PlayerShipStateEvent {}),
				),
			)
			.with_context(|| format!("while handling player state change event"))
			.with_context(|| format!("in ui scene `{}`", current_scene.as_ref().unwrap()))?;
		}

		let len = (*self.state).borrow().len();
		for i in 0..len {
			let event_arg = match (*self.state).borrow_mut().get_mut_by_idx(i).unwrap() {
				UiElement::Sprite(sprite) => {
					// Draw and update sprites
					sprite.step(&input, state);
					sprite.push_to_buffer(&input, state);
					let event = sprite.check_events(&input, state);

					match event {
						Event::None => None,

						Event::MouseClick => Some(Dynamic::from(MouseClickEvent {
							down: true,
							element: sprite.name.clone(),
						})),

						Event::MouseRelease => Some(Dynamic::from(MouseClickEvent {
							down: false,
							element: sprite.name.clone(),
						})),

						Event::MouseHover => Some(Dynamic::from(MouseHoverEvent {
							enter: true,
							element: sprite.name.clone(),
						})),

						Event::MouseUnhover => Some(Dynamic::from(MouseHoverEvent {
							enter: false,
							element: sprite.name.clone(),
						})),
					}
				}

				UiElement::RadialBar(x) => {
					// Draw and update radialbar
					x.step(&input, state);
					x.push_to_buffer(&input, state);
					None
				}

				UiElement::Text(..) => None,
			};

			if let Some(event_arg) = event_arg {
				rhai_error_to_anyhow(
					self.engine.call_fn(
						&mut self.scope,
						ct.get_config()
							.ui_scenes
							.get(current_scene.as_ref().unwrap().as_str())
							.unwrap(),
						"event",
						(State::new(input.clone()), event_arg.clone()),
					),
				)
				.with_context(|| format!("while handling event `{:?}`", event_arg))
				.with_context(|| format!("in ui scene `{}`", current_scene.as_ref().unwrap()))?;
			}
		}

		return Ok(());
	}
}

// API
impl UiScriptExecutor {
	pub fn register_api(
		ct_src: Arc<Content>,
		font_src: Rc<RefCell<FontSystem>>,
		s: Rc<RefCell<UiState>>,
		engine: &mut Engine,
	) {
		// Utilities
		{
			let c = s.clone();
			engine.register_fn("go_to_scene", move |scene: ImmutableString| {
				let mut ui_state = c.borrow_mut();
				ui_state.set_scene(scene);
			});

			let c = s.clone();
			engine.register_fn("conf_set_phys", move |b: bool| {
				let mut ui_state = c.borrow_mut();
				ui_state.config.show_phys = b;
			});

			let c = s.clone();
			engine.register_fn("conf_set_starfield", move |b: bool| {
				let mut ui_state = c.borrow_mut();
				ui_state.config.show_starfield = b;
			});
		}

		// Sprites
		{
			let c = s.clone();
			let ct = ct_src.clone();
			engine.register_fn(
				"add_sprite",
				move |name: ImmutableString, sprite: ImmutableString, rect: Rect| {
					let mut ui_state = c.borrow_mut();
					let len = ui_state.len();

					let sprite_handle = ct.get_sprite_handle(sprite.as_str());
					if sprite_handle.is_none() {
						error!("made a sprite using an invalid source `{sprite}`");
						return;
					}

					ui_state.names.insert(name.clone(), len);
					ui_state.elements.push(UiElement::Sprite(Sprite::new(
						&ct,
						name.clone(),
						sprite_handle.unwrap(),
						rect,
					)));
				},
			);

			let c = s.clone();
			let ct = ct_src.clone();
			engine.register_fn(
				"sprite_set_mask",
				move |name: ImmutableString, mask: ImmutableString| {
					let mut ui_state = c.borrow_mut();

					match ui_state.get_mut_by_name(&name) {
						Some(UiElement::Sprite(x)) => {
							let m = ct.get_sprite_handle(mask.as_str());
							if m.is_none() {
								error!("called `set_sprite_mask` with an invalid mask `{mask}`");
								return;
							}
							x.set_mask(m)
						}

						_ => {
							error!("called `set_sprite_mask` on an invalid name `{name}`")
						}
					}
				},
			);

			let c = s.clone();
			let ct = ct_src.clone();
			engine.register_fn(
				"sprite_take_edge",
				move |name: ImmutableString, edge_name: ImmutableString, duration: f32| {
					let mut ui_state = c.borrow_mut();

					match ui_state.get_mut_by_name(&name) {
						Some(UiElement::Sprite(x)) => {
							let sprite_handle = x.anim.get_sprite();
							let sprite = &ct.get_sprite(sprite_handle);

							let edge = resolve_edge_as_edge(edge_name.as_str(), duration, |x| {
								sprite.get_section_handle_by_name(x)
							});
							let edge = match edge {
								Err(_) => {
									error!(
									"called `sprite_take_edge` on an invalid edge `{}` on sprite `{}`",
									edge_name, sprite.name
								);
									return;
								}
								Ok(s) => s,
							};

							x.anim.jump_to(&ct, edge);
						}
						_ => {
							error!("called `sprite_take_edge` on an invalid name `{name}`")
						}
					}
				},
			);
		}

		// Textboxes
		{
			let c = s.clone();
			let font = font_src.clone();
			engine.register_fn(
				"add_textbox",
				// TODO: fix ugly spaces
				move |name: ImmutableString,
				      font_size: f32,
				      line_height: f32,
				      rect: Rect,
				      color: Color| {
					let mut ui_state = c.borrow_mut();
					let len = ui_state.len();

					ui_state.names.insert(name.clone(), len);
					ui_state.elements.push(UiElement::Text(TextBox::new(
						&mut font.borrow_mut(),
						name.clone(),
						font_size,
						line_height,
						rect,
						color,
					)));
				},
			);

			let c = s.clone();
			let font = font_src.clone();
			engine.register_fn(
				"textbox_set_text",
				move |name: ImmutableString, text: ImmutableString| {
					let mut ui_state = c.borrow_mut();
					match ui_state.get_mut_by_name(&name) {
						Some(UiElement::Text(x)) => {
							x.set_text(&mut font.borrow_mut(), text.as_str())
						}
						_ => {
							error!("called `textbox_set_text` on an invalid name `{name}`")
						}
					}
				},
			);

			let c = s.clone();
			let font = font_src.clone();
			engine.register_fn("textbox_align_left", move |name: ImmutableString| {
				let mut ui_state = c.borrow_mut();
				match ui_state.get_mut_by_name(&name) {
					Some(UiElement::Text(x)) => x.set_align(&mut font.borrow_mut(), Align::Left),
					_ => {
						error!("called `textbox_align_left` on an invalid name `{name}`")
					}
				}
			});

			let c = s.clone();
			let font = font_src.clone();
			engine.register_fn("textbox_align_right", move |name: ImmutableString| {
				let mut ui_state = c.borrow_mut();
				match ui_state.get_mut_by_name(&name) {
					Some(UiElement::Text(x)) => x.set_align(&mut font.borrow_mut(), Align::Right),
					_ => {
						error!("called `textbox_align_right` on an invalid name `{name}`")
					}
				}
			});

			let c = s.clone();
			let font = font_src.clone();
			engine.register_fn("textbox_align_justify", move |name: ImmutableString| {
				let mut ui_state = c.borrow_mut();
				match ui_state.get_mut_by_name(&name) {
					Some(UiElement::Text(x)) => {
						x.set_align(&mut font.borrow_mut(), Align::Justified)
					}
					_ => {
						error!("called `textbox_align_justify` on an invalid name `{name}`")
					}
				}
			});

			let c = s.clone();
			let font = font_src.clone();
			engine.register_fn("textbox_align_center", move |name: ImmutableString| {
				let mut ui_state = c.borrow_mut();
				match ui_state.get_mut_by_name(&name) {
					Some(UiElement::Text(x)) => x.set_align(&mut font.borrow_mut(), Align::Center),
					_ => {
						error!("called `textbox_align_center` on an invalid name `{name}`")
					}
				}
			});

			let c = s.clone();
			let font = font_src.clone();
			engine.register_fn("textbox_weight_bold", move |name: ImmutableString| {
				let mut ui_state = c.borrow_mut();
				match ui_state.get_mut_by_name(&name) {
					Some(UiElement::Text(x)) => x.set_weight(&mut font.borrow_mut(), Weight::BOLD),
					_ => {
						error!("called `textbox_weight_bold` on an invalid name `{name}`")
					}
				}
			});

			let c = s.clone();
			let font = font_src.clone();
			engine.register_fn("textbox_weight_normal", move |name: ImmutableString| {
				let mut ui_state = c.borrow_mut();
				match ui_state.get_mut_by_name(&name) {
					Some(UiElement::Text(x)) => {
						x.set_weight(&mut font.borrow_mut(), Weight::NORMAL)
					}
					_ => {
						error!("called `textbox_weight_normal` on an invalid name `{name}`")
					}
				}
			});

			let c = s.clone();
			let font = font_src.clone();
			engine.register_fn("textbox_font_serif", move |name: ImmutableString| {
				let mut ui_state = c.borrow_mut();
				match ui_state.get_mut_by_name(&name) {
					Some(UiElement::Text(x)) => {
						x.set_font(&mut font.borrow_mut(), FamilyOwned::Serif)
					}
					_ => {
						error!("called `textbox_font_serif` on an invalid name `{name}`")
					}
				}
			});

			let c = s.clone();
			let font = font_src.clone();
			engine.register_fn("textbox_font_sans", move |name: ImmutableString| {
				let mut ui_state = c.borrow_mut();
				match ui_state.get_mut_by_name(&name) {
					Some(UiElement::Text(x)) => {
						x.set_font(&mut font.borrow_mut(), FamilyOwned::SansSerif)
					}
					_ => {
						error!("called `textbox_font_sans` on an invalid name `{name}`")
					}
				}
			});

			let c = s.clone();
			let font = font_src.clone();
			engine.register_fn("textbox_font_mono", move |name: ImmutableString| {
				let mut ui_state = c.borrow_mut();
				match ui_state.get_mut_by_name(&name) {
					Some(UiElement::Text(x)) => {
						x.set_font(&mut font.borrow_mut(), FamilyOwned::Monospace)
					}
					_ => {
						error!("called `textbox_font_mono` on an invalid name `{name}`")
					}
				}
			});

			let c = s.clone();
			let font = font_src.clone();
			engine.register_fn("textbox_style_normal", move |name: ImmutableString| {
				let mut ui_state = c.borrow_mut();
				match ui_state.get_mut_by_name(&name) {
					Some(UiElement::Text(x)) => x.set_style(&mut font.borrow_mut(), Style::Normal),
					_ => {
						error!("called `textbox_style_normal` on an invalid name `{name}`")
					}
				}
			});

			let c = s.clone();
			let font = font_src.clone();
			engine.register_fn("textbox_style_italic", move |name: ImmutableString| {
				let mut ui_state = c.borrow_mut();
				match ui_state.get_mut_by_name(&name) {
					Some(UiElement::Text(x)) => x.set_style(&mut font.borrow_mut(), Style::Italic),
					_ => {
						error!("called `textbox_style_italic` on an invalid name `{name}`")
					}
				}
			});
		}

		// Radialbars
		{
			let c = s.clone();
			engine.register_fn(
				"add_radialbar",
				// TODO: fix ugly spaces
				move |name: ImmutableString, stroke: f32, color: Color, rect: Rect| {
					let mut ui_state = c.borrow_mut();
					let len = ui_state.len();

					ui_state.names.insert(name.clone(), len);
					ui_state.elements.push(UiElement::RadialBar(RadialBar::new(
						name.clone(),
						stroke,
						color,
						rect,
						1.0,
					)));
				},
			);

			let c = s.clone();
			engine.register_fn(
				"radialbar_set_val",
				move |name: ImmutableString, val: f32| {
					let mut ui_state = c.borrow_mut();
					match ui_state.get_mut_by_name(&name) {
						Some(UiElement::RadialBar(x)) => x.set_val(val),
						_ => {
							error!("called `radialbar_set_val` on an invalid name `{name}`")
						}
					}
				},
			);
		}
	}
}
