use galactica_content::Content;
use glyphon::TextArea;
use log::{debug, error};
use rhai::ImmutableString;
use std::collections::HashMap;
use std::sync::Arc;
use winit::window::Window;

use super::util::{FpsIndicator, RadialBar, Sprite, TextBox};
use crate::{RenderInput, RenderState};

#[derive(Debug)]
pub enum UiElement {
	Sprite(Sprite),
	RadialBar(RadialBar),
	Text(TextBox),
}

#[derive(Clone, Debug)]
pub(crate) struct UiConfig {
	pub show_phys: bool,
	pub show_starfield: bool,
}

pub(crate) struct UiState {
	pub names: HashMap<ImmutableString, usize>,
	pub elements: Vec<UiElement>,

	pub ct: Arc<Content>,
	current_scene: Option<ImmutableString>,

	show_timings: bool,
	fps_indicator: FpsIndicator,

	pub config: UiConfig,
}
// TODO: remove this
unsafe impl Send for UiState {}
unsafe impl Sync for UiState {}

impl UiState {
	pub fn new(ct: Arc<Content>, state: &mut RenderState) -> Self {
		Self {
			ct,
			names: HashMap::new(),
			elements: Vec::new(),

			current_scene: None,
			show_timings: true,
			fps_indicator: FpsIndicator::new(state),
			config: UiConfig {
				show_phys: false,
				show_starfield: false,
			},
		}
	}

	pub fn clear(&mut self) {
		self.elements.clear();
		self.names.clear();
	}

	pub fn len(&self) -> usize {
		self.elements.len()
	}

	/*
	pub fn get_by_idx(&self, idx: usize) -> Option<&UiElement> {
		self.elements.get(idx)
	}

	pub fn get_by_name(&self, name: &ImmutableString) -> Option<&UiElement> {
		let idx = self.names.get(name);
		if idx.is_none() {
			return None;
		}
		self.get_by_idx(*idx.unwrap())
	}
	*/

	pub fn get_mut_by_idx(&mut self, idx: usize) -> Option<&mut UiElement> {
		self.elements.get_mut(idx)
	}

	pub fn get_mut_by_name(&mut self, name: &ImmutableString) -> Option<&mut UiElement> {
		let idx = self.names.get(name);
		if idx.is_none() {
			return None;
		}
		self.get_mut_by_idx(*idx.unwrap())
	}

	pub fn get_scene(&self) -> &Option<ImmutableString> {
		&self.current_scene
	}

	pub fn set_scene(&mut self, scene: ImmutableString) {
		if !self.ct.get_config().ui_scenes.contains_key(scene.as_str()) {
			error!("tried to switch to ui scene `{scene}`, which doesn't exist");
			return;
		}

		debug!("switching to {:?}", scene);
		self.current_scene = Some(scene);
	}

	pub fn step(&mut self, state: &mut RenderState, input: Arc<RenderInput>) {
		if self.show_timings {
			self.fps_indicator
				.step(&input, &mut state.text_font_system.borrow_mut());
		}
	}
}

impl<'a> UiState {
	pub fn get_textareas(&'a mut self, input: &RenderInput, window: &Window) -> Vec<TextArea<'a>> {
		let mut v = Vec::with_capacity(32);

		if self.current_scene.is_none() {
			return v;
		}

		if self.show_timings {
			v.push(self.fps_indicator.get_textarea(input, window))
		}

		for t in self.elements.iter() {
			match &t {
				UiElement::Text(x) => v.push(x.get_textarea(input, window)),
				_ => {}
			}
		}

		return v;
	}
}
