use glyphon::{
	cosmic_text::Align, Attrs, AttrsOwned, Buffer, Color, FamilyOwned, FontSystem, Metrics,
	Shaping, Style, TextArea, TextBounds, Weight,
};
use nalgebra::Vector2;
use rhai::ImmutableString;
use winit::window::Window;

use super::super::api::Rect;
use crate::{ui::api, RenderInput};

#[derive(Debug)]
pub struct TextBox {
	pub name: ImmutableString,

	text: String,
	justify: Align,
	rect: Rect,
	buffer: Buffer,
	color: api::Color,
	attrs: AttrsOwned,
}

impl TextBox {
	pub fn new(
		font: &mut FontSystem,
		name: ImmutableString,
		font_size: f32,
		line_height: f32,
		rect: Rect,
		color: api::Color,
	) -> Self {
		let mut buffer = Buffer::new(font, Metrics::new(font_size, line_height));

		// Do NOT apply UI scale here, that's only done when we make a TextArea
		buffer.set_size(font, rect.dim.x, rect.dim.y);

		Self {
			name,
			rect,
			buffer,
			color,
			justify: Align::Left,
			attrs: AttrsOwned::new(Attrs::new()),
			text: String::new(),
		}
	}

	fn reflow(&mut self, font: &mut FontSystem) {
		self.buffer
			.set_text(font, &self.text, self.attrs.as_attrs(), Shaping::Advanced);

		for l in &mut self.buffer.lines {
			l.set_align(Some(self.justify));
		}

		self.buffer.shape_until_scroll(font);
	}

	pub fn set_text(&mut self, font: &mut FontSystem, text: &str) {
		self.text.clear();
		self.text.push_str(text);
		self.reflow(font);
	}

	pub fn set_align(&mut self, font: &mut FontSystem, align: Align) {
		self.justify = align;
		self.reflow(font);
	}

	pub fn set_weight(&mut self, font: &mut FontSystem, weight: Weight) {
		self.attrs.weight = weight;
		self.reflow(font);
	}

	pub fn set_font(&mut self, font: &mut FontSystem, family: FamilyOwned) {
		self.attrs.family_owned = family;
		self.reflow(font);
	}

	pub fn set_style(&mut self, font: &mut FontSystem, style: Style) {
		self.attrs.style = style;
		self.reflow(font);
	}
}

impl<'a, 'b: 'a> TextBox {
	pub fn get_textarea(&'b self, input: &RenderInput, window: &Window) -> TextArea<'a> {
		let rect = self
			.rect
			.to_centered(window, input.ct.get_config().ui_scale);

		// Glypon works with physical pixels, so we must do some conversion
		let fac = window.scale_factor() as f32;
		let corner_ne = Vector2::new(
			(rect.pos.x - rect.dim.x / 2.0) * fac + window.inner_size().width as f32 / 2.0,
			window.inner_size().height as f32 / 2.0 - (rect.pos.y * fac + rect.dim.y / 2.0),
		);
		let corner_sw = corner_ne + rect.dim * fac;
		let c = self.color.as_array_u8();

		TextArea {
			buffer: &self.buffer,
			top: corner_ne.y,
			left: corner_ne.x,
			scale: input.ct.get_config().ui_scale,
			bounds: TextBounds {
				top: (corner_ne.y) as i32,
				bottom: (corner_sw.y) as i32,
				left: (corner_ne.x) as i32,
				right: (corner_sw.x) as i32,
			},
			default_color: Color::rgba(c[0], c[1], c[2], c[3]),
		}
	}
}
