//! Provides a snapshot of one frame of a physics simulation

use std::collections::HashMap;

use super::{
	objects::{PhysEffect, PhysProjectile, PhysShip},
	PhysSimShipHandle,
};
use rapier2d::dynamics::RigidBody;

/// A snapshot of one frame of a physics simulation
#[derive(Debug)]
pub struct PhysImage {
	/// The ships in this frame
	pub(crate) ships: Vec<PhysShipImage>,

	/// Map ship handles to indices in ships
	pub(crate) ship_map: HashMap<PhysSimShipHandle, usize>,

	/// The projectiles in this frame
	pub(crate) projectiles: Vec<PhysProjectileImage>,

	/// The effects in this frame
	pub(crate) effects: Vec<PhysEffectImage>,
}

impl PhysImage {
	/// Create an empty simulation image
	pub fn new() -> Self {
		Self {
			ships: Vec::new(),
			projectiles: Vec::new(),
			ship_map: HashMap::new(),
			effects: Vec::new(),
		}
	}

	/// Clear all buffers in this image
	pub fn clear(&mut self) {
		self.ship_map.clear();
		self.projectiles.clear();
		self.effects.clear();
		self.ships.clear();
	}

	/// Iterate ships in this image
	pub fn iter_ships(&self) -> impl Iterator<Item = &PhysShipImage> {
		self.ships.iter()
	}

	/// Get a ship by its handle
	pub fn get_ship(&self, handle: &PhysSimShipHandle) -> Option<&PhysShipImage> {
		self.ship_map.get(handle).map(|x| &self.ships[*x])
	}

	/// Iterate projectiles in this image
	pub fn iter_projectiles(&self) -> impl Iterator<Item = &PhysProjectileImage> {
		self.projectiles.iter()
	}

	/// Iterate effects in this image
	pub fn iter_effects(&self) -> impl Iterator<Item = &PhysEffectImage> {
		self.effects.iter()
	}
}

/// A snapshot of a ship's state
#[derive(Debug)]
pub struct PhysShipImage {
	/// The ship's data
	pub ship: PhysShip,

	/// The ship's rigidbody
	pub rigidbody: RigidBody,
}

/// a snapshot of a projectile's state
#[derive(Debug)]
pub struct PhysProjectileImage {
	/// The projectile's data
	pub projectile: PhysProjectile,

	/// The projectile's rigidbody
	pub rigidbody: RigidBody,
}

/// a snapshot of a projectile's state
#[derive(Debug)]
pub struct PhysEffectImage {
	/// The effect's data
	pub effect: PhysEffect,

	/// The effect's rigidbody
	pub rigidbody: RigidBody,
}

impl PhysEffectImage {
	/// Get this effect's fade value
	pub fn get_fade(&self) -> f32 {
		let f = self.effect.fade;
		let l = self.effect.remaining_lifetime();
		return 1f32.min(l / f);
	}
}
